
/************************* SGA.c ; Single variable tuning  *************************/
#include <stdio.h>
#include <math.h>

#define MaxPop 30                            /* Maximum population size */
#define MaxString 31                          /* Maximum  string */

typedef enum {False=0, True} Boolean;           
typedef Boolean chromosome[MaxString];
typedef struct { chromosome chrom;               
		 float      x;                        /* ü 10  */
		 float      fitness;                  /* յ(Լ) */
		 int        parent1, parent2, xsite; } Individual;    
        /* parent1, parent2 : θ õ population ȣ      xsite : clossover site */
typedef Individual population[MaxString];

population oldpop, newpop;
int        popsize, lchrom, gen, maxgen;             
float      pcross, pmutation, sumfitness;
int        nmutation, ncross, jrand;
float      agv, max, min;
float      oldrand[56];        /* 56 random value */
  /* oldpop : old population( )        newpop : new population(ο )
   popsize : population size       lchrom : chromosome length(ü )       gen : generation  
   maxgen : maximum generation     pcross : crossover probability      agv : average of fitness     
   pmutation : mutation probability    sumfitness : summation of fitness    
   nmutation : number of mutation    ncross : number of crossover    max : maximum of fitness
   min : minimum of fitness          jrand : n° rnadom value (n=154) */

double power(x,y)     /****** x^n  ***************************************************/
   float x;
   int y;
{
   return exp(y*log(x));
}

//================================================================================
//
// name:          f3deceptive
//
// function:      computes the value of 3-deceptive function
//
// parameters:    x............a binary string
//                n............the length of a string
//
// returns:       (float) sum_{i=0}^{n/3-1}{f3(x[3i],x[3i+1],x[3i+2]}, where 
//                f3(0,0,0)=0.9, f3(0,0,1)=f3(0,1,0)=f3(1,0,0)=0.8,
//                f3(0,1,1)=f3(1,0,1)=f3(1,1,0)=0, and f3(1,1,1)=1
//
// ================================================================================
float objfunc(chrom)     /****** object function : f(x) = f3deceptive( chromosome ) *********/
   chromosome chrom;
{
	float temp1 = 0, temp2 = 0, result = 0;
	int i = 0, j = 0;
	
	for ( i = 1 ; i < lchrom ; i = i+3 )
	{
		for ( j = 0; j < 3 ; j++ )
		{
			if ( chrom[i+j] == True ) temp1++;
		}
		if ( temp1 == 3 ) temp2+=1;
		else if ( temp1 == 2 ) temp2+=0;
		else if ( temp1 == 1 ) temp2+=0.8;
		else if ( temp1 == 0 ) temp2+=0.9;

		result+=temp2;
		temp1 = 0;
		temp2 = 0;
	}
   return result;        /* xn  */
}

float decode(chrom, lbits)     /****** 2 10 ȯ *********************************/
  chromosome chrom;            /* chrom = oldpop[j].chrom */
  int lbits;                       /* lbits = lchrom */
{
  int j;
  float accum=0.0, powerof2=1.0;

  for(j=1; j<=lbits; j++) {  
  /* accum=chrom[1]*1+chrom[2]*2+chrom[3]*22+chrom[4]*23 */	  
          if (chrom[j]==True) accum+=powerof2; 
	powerof2 *= 2.0;
  }
    return accum;                /* 10   */
}

statistics(pop)     /****** ڿ()  (ִ밪, ּҰ, հ ) ************/
  population pop;                 /* pop = oldpop or newpop */
{
  int j;

  sumfitness = min = max = pop[1].fitness;
  for(j=2; j<=popsize; j++) {
	 sumfitness += pop[j].fitness;                   /* summation of fitness */
	 if(pop[j].fitness > max) max=pop[j].fitness;     /* maximum of fitness */
	 if(pop[j].fitness < min) min=pop[j].fitness;      /* minimum of fitness */
  }
  agv = sumfitness/popsize;                             /* average of fitness */
}

initdata()     /****** GAs parameter ʱⰪ , randomԼ  ************************/
{
   char ch;
   int  j;

   //clrscr();                       /* clear text mode window */
   printf("*******************************\n");
   printf("*  A Simple Genetic Algorithm *\n");
   printf("*  (c) David Edward Goldberg  *\n");
   printf("*  All Right Reserved         *\n");
   printf("*******************************\n");
   printf("**** SGA Data Entry and Initialization ****\n");     /* parameter ʱⰪ  */
   /* printf("Enter the Population Size ->\t");    scanf("%d",&popsize);*/
       popsize=30;
   /* printf("Enter chromosome length ->\t");    scanf("%d",&lchrom);*/
       lchrom=30;
   /* printf("Enter max. generations ->\t");      scanf("%d",&maxgen);*/
       maxgen=100;
   /* printf("Enter crossover probability ->\t");  scanf("%f",&pcross);*/
       pcross=0.6;
   /* printf("Enter mutation probability ->\t");   scanf("%f",&pmutation);*/
       pmutation=0.033;
   randomize();                   /*    randomԼ  */
   nmutation=ncross=0.0;          /* mutation, crossover ȸ 0 ʱȭ */
}

initreport()     /****** ʱ   (parameter ʱⰪ  Initial population ) ************/
{
   printf("*******************************\n");
   printf("*  A Simple Genetic Algorithm *\n");
   printf("*  (c) David Edward Goldberg  *\n");
   printf("*  All Right Reserved         *\n");
   printf("*******************************\n");
   printf(" #### SGA parameters ####\n\n");
   printf("Population Size -> %d\n",popsize);
   printf("Chromosome length -> %d\n",lchrom);
   printf("Maximum # of generations -> %d\n",maxgen);
   printf("Crossover probability -> %f\n",pcross);
   printf("Mutation probability -> %f\n\n",pmutation);
   printf(" ### Initial Generation Statistics ###\n");
   printf(" Initial population maximum fitness -> %f\n",max);
   printf(" Initial population average fitness -> %f\n",agv);
   printf(" Initial population minimum fitness -> %f\n",min);
   printf(" Initial population sum of fitness -> %f\n",sumfitness);
   getch();
}

advance_random()     /****** random value  ( random  )******************/
{
  int   j1;
  float new_random;

  for(j1=1; j1<=24; j1++) {        /* oldrand[1]oldrand[24]  */
	new_random = oldrand[j1] - oldrand[j1+31];
	if(new_random < 0.0) new_random += 1.0;
	oldrand[j1]=new_random;
  }
  for(j1=25; j1<=55; j1++) {       /* oldrand[25]oldrand[55]  */
	new_random = oldrand[j1] - oldrand[j1-24];
	if(new_random < 0.0) new_random += 1.0;
	oldrand[j1]=new_random;
  }
}

warmup_random(random_seed)     /****** 1 random value  ****************************/
	float random_seed;       /* random_seed = randomseed = Լ ʱⰪ */
{
   int  j1, ii;
   float new_random, prev_random;
        /* new_random : ο  ,    prev_random :    */

   oldrand[55] = random_seed;     /* random value   ʱȭ */
   new_random = 1.0e-9;          /* ʱȭ( ) */         
   prev_random = random_seed;    /* ʱȭ */
   for(j1=1; j1<=54; j1++) {       /* random value  -  ڿ 54 */
	  ii = (21*j1) % 55;     /* 054  */
	  oldrand[ii]=new_random;       /* oldrand[0]oldrand[54] random value  */
	  new_random = prev_random - new_random;
	  if(new_random < 0.0)  new_random+=1.0;
	  prev_random=oldrand[ii];
   }
   advance_random();   advance_random();   advance_random(); /*  random   */
   jrand=0;
}

float g_random()     /****** random value  ********************************************/
{
  jrand++;                       /*  ڿ random value  Ǿ ȣ */
  if(jrand > 55) { jrand=1; advance_random(); } /* 55° random  õ 
                                                  jrand 1 setingϰ ڿ   */
  return oldrand[jrand];          /* jrand° random value  */
}

Boolean g_flip(probability)     /****** Ƿ 0, 1  (ü ) ***********************/
  float probability;
{
  float tmp_random;                                /* ʿ  */
  if(probability == 1.0) return True;                 /* 0 or 1  */
	else return (g_random() <= probability);    /* random value Ͽ  ϸ
                                                      1   0  */
}

initpop()     /****** population ʱȭ (ʱ  ) *********************************/
{            /*  random value(oldrand[1]oldrand[55]) ̿ */
   int j, j1;
   for(j=1; j<=popsize; j++) {                         /* population  */
	for(j1=1; j1<=lchrom; j1++) {                 /* population ü  */
	     oldpop[j].chrom[j1]=g_flip(0.5);        /* 0 or 1    : probability=0.5 */
	}
	oldpop[j].x=decode(oldpop[j].chrom,lchrom);    /* 2(ü) 10 ȯ */
	oldpop[j].fitness=objfunc(oldpop[j].chrom);
	oldpop[j].parent1 = oldpop[j].parent2 = oldpop[j].xsite = 0.0; /*  븦  ʱȭ*/
   }
}

initialize()     /****** program ʱȭ ۾ (parameter ʱȭ  ʱ  ) ***********/ 
{
   initdata();                     /* GAs parameter ʱⰪ   randomԼ  */
   initpop();                      /* population ʱȭ(ʱ  ) */
   statistics(oldpop);              /*     (ִ밪, ּҰ, հ ) */
   initreport();                   /* ʱ   */
}

writechrom(chrom, lchrom)     /******  ü ȭ鿡 (2) ******************/
   chromosome chrom;           /* chrom = oldpop[i].chrom  ü */
   int        lchrom;             /* ü  */
{
   int        j;
   for(j=lchrom; j>=1; j--)       /*  ִ  ݴ  */
      if(chrom[j]) printf("1"); else printf("0");
   printf("\t");
}

report(gen)     /*****      ü   ( ) *******/
  int  gen;
{
   int i;

  for(i=1; i<=popsize; i++) {                      
   printf("G_%d\t",gen); printf("num_%d\t",i);        /*   ü   */
   writechrom(oldpop[i].chrom,lchrom);
   printf("%e\t",oldpop[i].x); printf("%f\n",oldpop[i].fitness);
   printf("G_:%d\t",gen+1); printf("num_%d\t",i);     /* ο  ü   */
   writechrom(newpop[i].chrom,lchrom);
   printf("%e\t",newpop[i].x); printf("%f\n",newpop[i].fitness);
  }
}

new_report(gen)     /******  뿡  ü   ******************************/
  int  gen;
{
   int i;

  for(i=1; i<=popsize; i++) {
   printf("G_%d\t",gen); printf("num_%d\t",i);    /* generation population ȣ  */
   writechrom(newpop[i].chrom,lchrom);           /* ü (2) */ 
	printf("%e\t",newpop[i].x); printf("%f\n",newpop[i].fitness);   /*10  յ  */
  }
	printf("--------------------------------------------------------------------\n");
	printf("Maximun -> %f\t Minimum -> %f\t Average -> %f\n",max,min,agv);
	printf("### sum of fitness -> %f\n\n\n\n",sumfitness);
	getch();        /* յ ִ밪, ּҰ, հ, ü յ  */
}

select(popsize, sumfitness, pop)     /****** 귿 select Ŀ population  *************/
   int        popsize;
   float      sumfitness;
   population pop;                 /* pop = oldpop */
{
   float      rand, partsum;       /* partsum : κ յ  */
   int        j;

   partsum = 0.0; j = 0;
   rand = g_random() * sumfitness;   /* rand=random value*sumfitness(ü յ) : ġ */
   do {
	 j++;
	 partsum += pop[j].fitness;
      } while((partsum < rand) && (j != popsize));   /* partsum rand(ġ) ũ  */
   return j;                          /* õ population ȣ  */
}

Boolean mutation(alleleval, pmutation, nmutation)     /****** mutation   **************/
   Boolean alleleval;              /* alleleval = parent1[j], parent2[j] */
   float   pmutation;
   int     nmutation;
{
   Boolean	   mutate;

   mutate = g_flip(pmutation);     /* mutation ɼ 0.033   0, 1  */
   if(mutate) {
	nmutation++;             /* mutation  ȸ */
	return !alleleval;          /*   */
   }
   else  return alleleval;           /* ״  */
}

crossover(parent1, parent2, child1, child2, lchrom, ncross, nmutation,     
         jcross, pcross, pmutation)     /****** clossover   ***************************/
  chromosome parent1, parent2, child1, child2;    /* parent1=oldpop[mate1].chrom */
                /* parent2=oldpop[mate2],  child1=newpop[j].chrom,  child2=newpop[j+1].chrom */
  int        lchrom, ncross, nmutation, jcross;     /* jcross : crossover site */
  float      pcross, pmutation;
{
  int        j;

  if(g_flip(pcross)) {         /* pcross = 0.6  crossover ɼ 0.6  0, 1 */
      jcross = rnd(1,lchrom-1);       /* crossover site  */
      ncross++;
   }
  else   jcross = lchrom;            /* crossover   */
  for(j=1; j<=jcross; j++) {          /* mutation   */
      child1[j] = mutation(parent1[j], pmutation, nmutation);
      child2[j] = mutation(parent2[j], pmutation, nmutation);
  }
  if(jcross != lchrom) {             /* crossover  mutation   */
     for(j=jcross+1; j<=lchrom; j++) {
	  child1[j] = mutation(parent2[j], pmutation, nmutation);
	  child2[j] = mutation(parent1[j], pmutation, nmutation);
     }
  }
}

generation()     /******    ****************************************************/
{
  int  j=1, mate1, mate2, jcross, i=0;       /*mate1, mate2 : θ õ population ȣ */
                                          /* jcross : crossover site */
  do {
       mate1 = select(popsize, sumfitness, oldpop);    /* 귿 selectĿ   */
       mate2 = select(popsize, sumfitness, oldpop);
       crossover(oldpop[mate1].chrom, oldpop[mate2].chrom,
			 newpop[j].chrom, newpop[j+1].chrom,
     		 lchrom, ncross, nmutation, jcross, pcross, pmutation); 
       /* newpop[j] = child1 */
       newpop[j].x = decode(newpop[j].chrom,lchrom);
       newpop[j].fitness = objfunc(newpop[j].chrom);
       newpop[j].parent1 = mate1;      /* õ θ population ȣ */
       newpop[j].parent2 = mate2;      
       newpop[j].xsite = jcross;         /* crossover site */
       /* newpop[j+1] = child2 */
       newpop[j+1].x = decode(newpop[j+1].chrom,lchrom);
       newpop[j+1].fitness = objfunc(newpop[j+1].chrom);
       newpop[j+1].parent1 = mate1;
       newpop[j+1].parent2 = mate2;
       newpop[j+1].xsite = jcross;
       j+=2;
   } while(j <= popsize);
}

pause(pauselength)     /******   Լ -   **************************/
   int pauselength;
{
   int   maxpause = 2500, j, j1;
   float x;

   for(j=1; j<pauselength; j++)
     for(j1=1; j1<maxpause; j1++) x = 0.0 + 1.0;
}

rnd(low, high)     /****** crossover site  ( ִ(lchrom-1), ּ(1)    ) **/
  int low, high;
{
   int i;
   if(low>high) i=low;
    else {
       i = floor(g_random()*(high-low+1) +low); /* ִ(lchrom-1), ּ(1)     */
       if(i>high) i=high;
     }
	return i;      /* clossover site */
}

randomize()     /****** ⺻ random value ʱȭ, random value  ************************/
{
  float randomseed;     /* randomԼ ʱⰪ */
  do {
        /* printf("### Enter the random number (0.0 ~ 1.0) ->\t");    scanf("%f",&randomseed);*/
	  randomseed=0.2;     /* ʱⰪ  */
	  } while((randomseed<0.0) || (randomseed>1.0));
   warmup_random(randomseed);      /* random value  */
}

new_to_old()     /****** buffer (     ) ***********************************/
{
 int i,j;

  for(i=1; i<=popsize; i++) {
     oldpop[i].x=newpop[i].x;
     oldpop[i].fitness=newpop[i].fitness;
     oldpop[i].parent1=newpop[i].parent1;
     oldpop[i].parent2=newpop[i].parent2;
     oldpop[i].xsite=newpop[i].xsite;
     for(j=1; j<=MaxString; j++) oldpop[i].chrom[j]=newpop[i].chrom[j];
  }
}

main() {
	gen=0;          /*  ʱȭ */
	initialize();      /* promgram ʱȭ(parameter, ramdomԼ,   ʱⰪ  */
	do {
		gen++;                   /*   */
		generation();             /*    */
		statistics(newpop);        /* ο 뿡    */
		new_report(gen);         /* ο    */
		new_to_old();            /* buffer(  غ) */
	 }  while(gen < maxgen);         /* ϴ  ŭ ݺ (maxgen=200) */
}
